include "include_test2.fbs";    // should be skipped

attribute "included_attribute";

namespace MyGame.OtherNameSpace;

enum FromInclude:long { IncludeVal, Foo = 17 }

struct Unused { unused: byte; }


